from multiprocessing import Process
import opendssdirect as dss
import numpy as np
import os

class DLF(object):

    def __init__(self):
        self.BMva = 100
        self.KvBase = 132.79
        self.phases = 3
        self.npts = 24
        self.step = 1
        self.i = 0

    def createckt(self, V_a, V_b, V_c, Angle_a, Angle_b, Angle_c, loadbus):
        # Create a new circuit for OpenDSS
        address = 'Address to write source file'+str(loadbus)+'/source.txt'
        fh = open(address, "w")
        fh.write("New Circuit.ckt bus1=SourceBus.1 angle={0} pu={1} basekV={2} phase=1 r1=0  x1=0.001  r0=0  x0=0.001\n".format(Angle_a, V_a, self.KvBase))
        fh.write("New Vsource.Source_2 bus1=SourceBus.2 angle={0} pu={1} basekV={2} phase=1 r1=0  x1=0.001  r0=0  x0=0.001\n".format(Angle_b, V_b, self.KvBase))
        fh.write("New Vsource.Source_3 bus1=SourceBus.3 angle={0} pu={1} basekV={2} phase=1 r1=0  x1=0.001  r0=0  x0=0.001\n".format(Angle_c, V_c, self.KvBase))
        fh.close()
        return [loadbus]

    def LFSETUP(self):
        # Returns the error matrix comprising of each phase power and voltage
        r = dss.run_command(r'Address to OpenDss circuit file')
        dss.run_command('set mode=daily')
        dss.run_command('set stepsize=1h')
        dss.run_command('set number=1')

    def COMPILE(self, system, loadbus):
        if loadbus == 4:
            if system == 'ckt24':
                r = dss.run_command('Address to OpenDss circuit file')
                print(r)
            elif system == 'ckt7':
                s = dss.run_command('Address to OpenDss circuit file')
                print(s)
        elif loadbus == 5:
            if system == 'ckt24':
                dss.run_command('Address to OpenDss circuit file')
            elif system == 'ckt7':
                dss.run_command('Address to OpenDss circuit file')
        elif loadbus == 7:
            if system == 'ckt24':
                dss.run_command('Address to OpenDss circuit file')
            elif system == 'ckt7':
                dss.run_command('Address to OpenDss circuit file')

    def LFTIMESERIES(self, system):
        s = dss.run_command('solve')
        if system == 'ckt24':
            dss.Circuit.SetActiveElement('Transformer.SUBXFMR')
            S = dss.CktElement.Powers()[:6]  # Fetch the high side of transformer power which is source power
        elif system == 'ckt7':
            dss.Circuit.SetActiveElement('Transformer.Sub1')
            S1 = dss.CktElement.Powers()[:6]
            dss.Circuit.SetActiveElement('Transformer.Sub2')
            S2 = dss.CktElement.Powers()[:6]
            dss.Circuit.SetActiveElement('Transformer.Sub3')
            S3 = dss.CktElement.Powers()[:6]
            S = [sum(pair) for pair in zip(S1, S2, S3)]
        # Fetch the high side of transformer voltages which are source voltages
        V = dss.CktElement.Voltages()[:6]
        V_real = np.array(V[::2])
        V_imag = np.array(V[1::2])
        V_abs = np.absolute((V_real + 1j * V_imag) / (1000*self.KvBase))

        # Divide the P in KW by 1000 (MW) and base MVA
        P = np.array(S[::2]) / (1000*self.BMva)
        Q = np.array(S[1::2]) / (1000*self.BMva)
        Cosim_error = np.concatenate((P, Q, V_abs), axis=0)
        error_matrix = {'Cosim_error': Cosim_error,
                        'Real_Power': P,
                        'Reactive_Power': Q,
                        'Interface_Voltage': V_abs}
        return error_matrix

    def main(self):
        a = DLF()
        b = a.createckt(5)
        a.COMPILE(*b, 'ckt7')
        c = a.LFTIMESERIES('ckt7')
        print(c)

if __name__ == '__main__':
    DLF.main('distributionsystem')