import os
import numpy as np
import collections
import cmath
import math
from Functions import arraymXn, complexno, admittancedia, admittanceoffdia, admittance, threephasearr, threephasetoseq
from Line_Series_ImpedanceData import seriesdata
from Line_Shunt_AdmittanceData import shuntdata

class SDTPF(object):
    # Base MVA
    BMva = 100
    # Base Current
    BCurrent = 251.02
    Phases = 3
    # Transformation vector in sequence matrix
    a = -0.5 + 0.8660 * (cmath.sqrt(-1))
    asqr = a ** 2
    acub = a ** 3
    # In standard transformers high voltage side leads low voltage side by 30
    transangle = 30

    # Sequence transformation matrices
    # Power variant sequence transformation matrix
    T1 = (1 / 3) * np.array([[1, 1, 1], [1, a, asqr], [1, asqr, a]])
    T2 = np.array([[1, 1, 1], [1, asqr, a], [1, a, asqr]])
    # Power invariant sequence transformation matrix
    T3 = (1 / (cmath.sqrt(3))) * np.array([[1, 1, 1], [1, a, asqr], [1, asqr, a]])
    T4 = (1 / (cmath.sqrt(3))) * np.array([[1, 1, 1], [1, asqr, a], [1, a, asqr]])
    def __init__(self, system):
        # Path to folder
        self.count = 0
        self.error = 0
        self.V3ph = 0
        self.Vseq = 0
        self.Vseq_PV = 0
        self.Vseq_final = 0
        self.V3ph_final = 0
        self.I3phasp = 0
        self.Iseqasp = 0
        self.Iseqasp_PV = 0
        self.Va = 0
        self.Vb = 0
        self.Vc = 0
        self.V0 = 0
        self.V1 = 0
        self.V2 = 0
        self.V0_PV = 0
        self.V1_PV = 0
        self.V2_PV = 0
        self.system = system
        self.filepath = os.path.dirname(os.path.abspath(__file__))
        self.get_data()

    # get_data reads the system data for transmission model based on function parameter (e.g: SDTPF(9) call reads IEEE 9
    # bus data from the respective folder)
    def get_data(self):
        self.gendata = np.loadtxt('Generator_Data.csv', dtype=np.float, delimiter=',', skiprows=1)
        self.busdata = np.loadtxt('Bus_Data.csv', dtype=np.float, delimiter=',', skiprows=1)
        self.transdata = np.loadtxt('Transformer_Data.csv', dtype=np.float, delimiter=',', skiprows=1)
        self.linedata = np.loadtxt('Line_Data.csv', dtype=np.float, delimiter=',', skiprows=1)
        self.branchdata = np.concatenate((self.transdata, self.linedata), axis=0)

    #compiledata formulates the initial three phase and sequence voltages and sequence current injections for unbalanced model
    def compiledata(self):
        cd = {}
        self.error = 1
        self.count = 0
        i = 1
        j = 1
        k = 1
        self.BranchFrom = self.branchdata[:, 0]
        self.BranchTo = self.branchdata[:, 1]
        bus = self.busdata[:, 0]
        nbus = np.shape(bus)[0]
        # No. of branches
        nl = np.shape(self.BranchFrom)[0]
        # Assuming first column is low voltage side
        transfrom = self.transdata[:, 0]
        # Number of transformers
        transno = np.shape(self.transdata)[0]
        # Number of lines
        lineno = np.shape(self.linedata)[0]
        linefrom = self.linedata[:, 0]
        lineto = self.linedata[:, 1]
        # Finds the number of times bus number is repeated in FromBranch
        FromBranch = collections.Counter(linefrom)
        # Finds the number of times bus number is repeated in ToBranch
        ToBranch = collections.Counter(lineto)
        # Type of Bus 3-Slack, 2-PV, 1-PQ
        typ = self.busdata[:, 1]
        Pg = self.busdata[:, 4] / self.BMva
        Qg = self.busdata[:, 5] / self.BMva
        Pl = self.busdata[:, 6] / (self.BMva * 3)
        Ql = self.busdata[:, 7] / (self.BMva * 3)
        Pload = self.busdata[:, 6]
        Qload = self.busdata[:, 7]
        # No. of PQ buses
        npq = np.count_nonzero(typ == 1)
        # No. of PV buses
        npv = np.shape(self.gendata)[0]
        pq = np.where(typ == 1)
        pq = pq[0]
        pv = np.where(typ == 2)
        pv = pv[0]
        # Initial Voltages
        V = self.busdata[:, 2]
        angle_deg = self.busdata[:, 3]
        # Transformer Taps
        Tap = self.branchdata[:, 2]
        R0 = self.branchdata[:, 3]
        X0 = self.branchdata[:, 4]
        R1 = self.branchdata[:, 6]
        X1 = self.branchdata[:, 7]
        R2 = self.branchdata[:, 9]
        X2 = self.branchdata[:, 10]
        B0 = self.branchdata[:, 5]
        B1 = self.branchdata[:, 8]
        B2 = self.branchdata[:, 11]

        [z0, z1, z2] = [complexno(R0, X0), complexno(R1, X1), complexno(R2, X2)]
        [s0, self.s1, s2] = [complexno(0, B0), complexno(0, B1), complexno(0, B2)]
        [y0, y1, y2] = [np.reciprocal(z0), np.reciprocal(z1), np.reciprocal(z2)]
        [Y0, Y1, Y2] = [arraymXn(nbus, nbus), arraymXn(nbus, nbus), arraymXn(nbus, nbus)]

        self.V3ph = arraymXn(self.Phases, nbus)
        self.Vseq = arraymXn(self.Phases, nbus)
        self.Vseq_PV = arraymXn(self.Phases, nbus)
        self.Vseq_final = arraymXn(self.Phases, nbus)
        self.V3ph_final = arraymXn(self.Phases, nbus)
        self.I3phasp = arraymXn(self.Phases, nbus)
        self.Iseqasp = arraymXn(self.Phases, nbus)
        self.Iseqasp_PV = arraymXn(self.Phases, nbus)

        # Angle
        del_1 = np.real(arraymXn(nbus, 1))

        # Create Admittance Matrices (Y1 - Full, Y0, Y2 - diagonal only)
        Y0 = admittancedia(nbus, nl, i, j, y0, Tap, s0, Y0, self.BranchFrom, self.BranchTo)
        Y1 = admittance(nbus, nl, i, j, k, y1, Tap, self.s1, Y1, self.BranchFrom, self.BranchTo)
        Y2 = admittancedia(nbus, nl, i, j, y2, Tap, s2, Y2, self.BranchFrom, self.BranchTo)

        Y1_abs = np.absolute(Y1)
        Y1_ang = np.angle(Y1)
        Y1_rea = np.real(Y1)
        Y1_img = np.imag(Y1)

        Y0_mx = np.asmatrix(Y0)
        Y2_mx = np.asmatrix(Y2)
        Y0_inv_mx = np.linalg.inv(Y0_mx)
        Y2_inv_mx = np.linalg.inv(Y2_mx)

        # Negative and Zero sequence admittance matrices are diagonal because off diagonal elements (mutual coupling) are considered as current injections
        Y0_inv = np.asarray(Y0_inv_mx)
        Y2_inv = np.asarray(Y2_inv_mx)

        # Initial voltages in three phases
        self.Va = V * self.acub
        self.Vb = V * self.asqr
        self.Vc = V * self.a

        # Create an array with three phase voltages
        self.V3ph = threephasearr(nl, k, self.Va, self.Vb, self.Vc, self.V3ph)

        # Create sequence voltages from three phase voltages
        self.Vseq = threephasetoseq(nl, k, self.T1, self.V3ph, self.Vseq)
        self.Vseq_PV = threephasetoseq(nl, k, self.T3, self.V3ph, self.Vseq_PV)

        [self.V0, self.V1, self.V2] = [self.Vseq[0, :], self.Vseq[1, :], self.Vseq[2, :]]
        [self.V0_PV, self.V1_PV, self.V2_PV] = [self.Vseq_PV[0, :], self.Vseq_PV[1, :], self.Vseq_PV[2, :]]

        dIz0 = arraymXn(nl, nl)
        dIz1 = arraymXn(nl, nl)
        dIz2 = arraymXn(nl, nl)
        dIs0from = arraymXn(nl, 1)
        dIs1from = arraymXn(nl, 1)
        dIs2from = arraymXn(nl, 1)
        dIs0to = arraymXn(nl, 1)
        dIs1to = arraymXn(nl, 1)
        dIs2to = arraymXn(nl, 1)

        # Current injections for any sequence component depends on coupling with only other sequence admittances and sequence
        # voltages at the bus. For series injection dIm = ymn(Vn(i) - Vn(j) + yml(Vl(i) - V1(j)). For shunt injections
        # dIm = ymn*Vn(i) + yml*Vl(i) where m = n = l = 0, 1, 2 and m != n != l
        # Off-diagonal current injection in the series admittance matrix
        for k in range(transno, nl):
            linefrom = int(self.BranchFrom[k])
            lineto = int(self.BranchTo[k])
            line = 'Y012_' + str(linefrom) + str(lineto)
            Y012 = seriesdata[line]
            dIz0[linefrom - 1, lineto - 1] = Y012[0, 1] * (self.V1[linefrom - 1] - self.V1[lineto - 1]) + Y012[0, 2] * (self.V2[linefrom - 1] - self.V2[lineto - 1])
            dIz1[linefrom - 1, lineto - 1] = Y012[1, 0] * (self.V0[linefrom - 1] - self.V0[lineto - 1]) + Y012[1, 2] * (self.V2[linefrom - 1] - self.V2[lineto - 1])
            dIz2[linefrom - 1, lineto - 1] = Y012[2, 0] * (self.V0[linefrom - 1] - self.V0[lineto - 1]) + Y012[2, 1] * (self.V1[linefrom - 1] - self.V1[lineto - 1])

        # Off-diagonal current injection in the shunt admittance matrix
        for m in range(transno, nl):
            linefrom = int(self.BranchFrom[m])
            lineto = int(self.BranchTo[m])
            line = 'S012_' + str(linefrom) + str(lineto)
            S012 = shuntdata[line]
            dIs0from[linefrom - 1] = dIs0from[linefrom - 1] + S012[0, 1] * self.V1[linefrom - 1] + S012[0, 2] * self.V2[linefrom - 1]
            dIs1from[linefrom - 1] = dIs1from[linefrom - 1] + S012[1, 0] * self.V0[linefrom - 1] + S012[1, 2] * self.V2[linefrom - 1]
            dIs2from[linefrom - 1] = dIs2from[linefrom - 1] + S012[2, 0] * self.V0[linefrom - 1] + S012[2, 1] * self.V1[linefrom - 1]

            dIs0to[lineto - 1] = dIs0to[lineto - 1] + S012[0, 1] * self.V1[lineto - 1] + S012[0, 2] * self.V2[lineto - 1]
            dIs1to[lineto - 1] = dIs1to[lineto - 1] + S012[1, 0] * self.V0[lineto - 1] + S012[1, 2] * self.V2[lineto - 1]
            dIs2to[lineto - 1] = dIs2to[lineto - 1] + S012[2, 0] * self.V0[lineto - 1] + S012[2, 1] * self.V1[lineto - 1]

        # Shunt current injections at every bus get added up for positive, negative and zero sequence. If the bus no gets
        # repeated in BranchFrom or BranchTo, shunt injections is added those many times. The following block of code checks
        # for repetitions and divides them based on number of occurences.
        for BranchNo, count in FromBranch.items():
            if count != 1:
                dIs0from[int(BranchNo) - 1] = dIs0from[int(BranchNo) - 1] / count
                dIs1from[int(BranchNo) - 1] = dIs1from[int(BranchNo) - 1] / count
                dIs2from[int(BranchNo) - 1] = dIs2from[int(BranchNo) - 1] / count

        for BranchNo, count in ToBranch.items():
            if count != 1:
                dIs0to[int(BranchNo) - 1] = dIs0to[int(BranchNo) - 1] / count
                dIs1to[int(BranchNo) - 1] = dIs1to[int(BranchNo) - 1] / count
                dIs2to[int(BranchNo) - 1] = dIs2to[int(BranchNo) - 1] / count

        # Sequence current injections at each bus
        dI0from = arraymXn(nl, 1)
        dI1from = arraymXn(nl, 1)
        dI2from = arraymXn(nl, 1)
        dI0to = arraymXn(nl, 1)
        dI1to = arraymXn(nl, 1)
        dI2to = arraymXn(nl, 1)
        dS1 = arraymXn(nl, 1)


        for n in range(transno, nl):
            linefrom = int(self.BranchFrom[n])
            lineto = int(self.BranchTo[n])
            dI0from[linefrom - 1] = dI0from[linefrom - 1] + dIs0from[linefrom - 1] + dIz0[linefrom - 1, lineto - 1]
            dI1from[linefrom - 1] = dI1from[linefrom - 1] + dIs1from[linefrom - 1] + dIz1[linefrom - 1, lineto - 1]
            dI2from[linefrom - 1] = dI2from[linefrom - 1] + dIs2from[linefrom - 1] + dIz2[linefrom - 1, lineto - 1]

            dI0to[lineto - 1] = dI0to[lineto - 1] + dIs0to[lineto - 1] - dIz0[linefrom - 1, lineto - 1]
            dI1to[lineto - 1] = dI1to[lineto - 1] + dIs1to[lineto - 1] - dIz1[linefrom - 1, lineto - 1]
            dI2to[lineto - 1] = dI2to[lineto - 1] + dIs2to[lineto - 1] - dIz2[linefrom - 1, lineto - 1]

        dI0 = dI0from + dI0to
        dI1 = dI1from + dI1to
        dI2 = dI2from + dI2to

        dI1_conj = np.conj(dI1)

        for q in range(nl):
            dS1[q] = self.V1_PV[q] * dI1_conj[q]
        dP1 = np.real(dS1)
        dQ1 = np.imag(dS1)
        return [del_1, dP1, dQ1, dI0, dI1, dI2, nbus, npv, npq, nl, Pg, Qg, Pl, Ql, pq, pv, Y1, Y1_abs, Y1_ang, Y1_rea,
                Y1_img, Y0_inv, Y2_inv, typ, Pload, Qload]

    # threeseq_solve runs the power flow by building Jacobian matrix for positive sequence , linear formulation negative
    # and zero sequence
    def threeseq_solve(self, del_1, dP1, dQ1, dI0, dI1, dI2, nbus, npv, npq, nl, Pg, Qg, Pl, Ql, pq, pv, Y1, Y1_abs, Y1_ang,
                       Y1_rea, Y1_img, Y0_inv, Y2_inv, typ, Pload, Qload, Pl_update=None, Ql_update=None):
        k = 1

        Pl_a = np.copy(Pl)
        Pl_b = np.copy(Pl)
        Pl_c = np.copy(Pl)
        Ql_a = np.copy(Ql)
        Ql_b = np.copy(Ql)
        Ql_c = np.copy(Ql)

        if Pl_update is None:
            pass
        else:
            Pl_a[5] = Pl_update[0]
            Pl_b[5] = Pl_update[1]
            Pl_c[5] = Pl_update[2]

        if Ql_update is None:
            pass
        else:
            Ql_a[5] = Ql_update[0]
            Ql_b[5] = Ql_update[1]
            Ql_c[5] = Ql_update[2]


        Ssp_a = complexno(Pl_a, Ql_a)
        Ssp_b = complexno(Pl_b, Ql_b)
        Ssp_c = complexno(Pl_c, Ql_c)

        # Three phase currents from specified power
        Iasp = arraymXn(1, nl)
        Ibsp = arraymXn(1, nl)
        Icsp = arraymXn(1, nl)
        self.error = 1
        self.count = 0

        # While loop solves until error is within tolerance
        while self.error > 0.00001:
            for t in range(nbus):
                Iasp[:, t] = np.conj((Ssp_a[t] / self.Va[t]))
                Ibsp[:, t] = np.conj((Ssp_b[t] / self.Vb[t]))
                Icsp[:, t] = np.conj((Ssp_c[t] / self.Vc[t]))

            for k in range(nl):
                self.I3phasp[:, k] = np.array([Iasp[0, k], Ibsp[0, k], Icsp[0, k]])

            # Sequence current injections
            # Since the standard three phase to sequence transformation is power variant, Iseqasp_PV sequence elements
            # are transformed by using T3 three phase to sequence power invariant matrix. Any sequence elements used in
            # calculating power should use T3 matrix for transformation
            self.Iseqasp = threephasetoseq(nl, k, self.T1, self.I3phasp, self.Iseqasp)
            self.Iseqasp_PV = threephasetoseq(nl, k, self.T3, self.I3phasp, self.Iseqasp_PV)


            Iload_0sp = self.Iseqasp[0, :]
            Iload_1 = self.Iseqasp[1, :]
            Iload_2sp = self.Iseqasp[2, :]

            Iload_0sp_PV = self.Iseqasp_PV[0, :]
            Iload_1_PV = self.Iseqasp_PV[1, :]
            Iload_2sp_PV = self.Iseqasp_PV[2, :]

            if self.count != 0:
                self.V0_PV = self.V0 * cmath.sqrt(3)
                self.V1_PV = self.V1 * cmath.sqrt(3)
                self.V2_PV = self.V2 * cmath.sqrt(3)

            Sload_1sp = arraymXn(nl, 1)
            Sload_0sp = arraymXn(nl, 1)
            Sload_2sp = arraymXn(nl, 1)

            # Sequence current injections are converted to sequence powers
            for n in range(nbus):
                Sload_1sp[n] = self.V1_PV[n] * np.conj(Iload_1_PV[n])
                Sload_0sp[n] = self.V0_PV[n] * np.conj(Iload_0sp_PV[n])
                Sload_2sp[n] = self.V2_PV[n] * np.conj(Iload_2sp_PV[n])

            Pload_1sp = np.real(Sload_1sp)
            Qload_1sp = np.imag(Sload_1sp)

            # Final sequence specifications at any bus
            P1_sp = arraymXn(nl, 1)
            Q1_sp = arraymXn(nl, 1)
            I2_sp = arraymXn(nl, 1)
            I0_sp = arraymXn(nl, 1)
            V1_abs = np.absolute(self.V1)

            # Total sequence specifications
            for i in range(nbus):
                if typ[i] == 1:
                    P1_sp[i] = -(Pload_1sp[i] + dP1[i])
                    Q1_sp[i] = -(Qload_1sp[i] + dQ1[i])
                    I2_sp[i] = (-Iload_2sp[i] - dI2[i])
                    I0_sp[i] = (-Iload_0sp[i] - dI0[i])
                elif typ[i] == 2:
                    P1_sp[i] = Pg[i]
                    Q1_sp[i] = Qg[i]

            # Calculated powers for Newton-Raphson
            P1 = arraymXn(nl, 1)
            Ql = arraymXn(nl, 1)
            for i in range(nbus):
                for j in range(nbus):
                    if j != i:
                        P1[i] = P1[i] + V1_abs[i] * V1_abs[j] * Y1_abs[i, j] * math.cos((Y1_ang[i, j] + del_1[j] - del_1[i]))
                        Ql[i] = Ql[i] + V1_abs[i] * V1_abs[j] * Y1_abs[i, j] * math.sin((Y1_ang[i, j] + del_1[j] - del_1[i]))

                P1[i] = P1[i] + V1_abs[i] * V1_abs[i] * Y1_rea[i, i]
                Ql[i] = -(Ql[i] + (V1_abs[i] * V1_abs[i] * Y1_img[i, i]))

            deltaP = arraymXn(nbus, 1)
            deltaQ = arraymXn(nbus-npv, 1)

            # Mismatch Matrix
            M = arraymXn(2 * nbus - npv - 1, 1)

            ctr = 0
            for i in range(1, nbus):
                deltaP[i] = P1_sp[i] - P1[i]
                M[ctr] = deltaP[i]
                ctr = ctr + 1
            j = 0
            for i in range(nbus):
                if typ[i] == 1:
                    deltaQ[j] = Q1_sp[i] - Ql[i]
                    M[ctr] = deltaQ[j]
                    ctr = ctr + 1
                    j = j + 1

            # Form the Jacobian Matrix
            J = arraymXn(((2 * nbus) - 1 - npv), ((2 * nbus) - 1 - npv))

            # Jptheta - Derivative of real power injections with angles
            Jptheta = np.real(arraymXn(nbus - 1, nbus - 1))
            for i in range(nbus - 1):
                n = i + 1
                for k in range(nbus - 1):
                    m = k + 1
                    if m == n:
                        # Diagonal elements
                        for n in range(nbus):
                            if m != n:
                                Jptheta[i, k] = Jptheta[i, k] + V1_abs[m] * V1_abs[n] * Y1_abs[m, n] * math.sin(
                                    del_1[n] - del_1[m] + Y1_ang[m, n])
                    else:
                        Jptheta[k, i] = -V1_abs[m] * V1_abs[n] * Y1_abs[m, n] * math.sin(
                            del_1[n] - del_1[m] + Y1_ang[m, n])
                        Jptheta[i, k] = -V1_abs[m] * V1_abs[n] * Y1_abs[m, n] * math.sin(
                            del_1[m] - del_1[n] + Y1_ang[m, n])

            # Jpv - Derivative of Real Power injections with Voltage
            Jpv = arraymXn(nbus - 1, npq)
            for i in range(nbus - 1):
                m = i + 1
                for k in range(npq):
                    n = pq[k]
                    if n == m:
                        for n in range(nbus):
                            if m != n:
                                Jpv[i, k] = Jpv[i, k] + V1_abs[n] * Y1_abs[m, n] * math.cos(
                                    Y1_ang[m, n] + del_1[n] - del_1[m])
                        Jpv[i, k] = Jpv[i, k] + 2 * V1_abs[m] * Y1_rea[m, m]
                    else:
                        Jpv[i, k] = V1_abs[m] * Y1_abs[m, n] * math.cos(del_1[n] - del_1[m] + Y1_ang[m, n])

            # Jqtheta - Derivative of Reactive Power Injections with Angles
            Jqtheta = np.real(arraymXn(npq, nbus - 1))
            for i in range(npq):
                m = pq[i]
                for k in range(nbus - 1):
                    n = k + 1
                    if n == m:
                        for n in range(nbus):
                            if n != m:
                                Jqtheta[i, k] = Jqtheta[i, k] + V1_abs[m] * V1_abs[n] * Y1_abs[m, n] * math.cos(
                                    del_1[n] - del_1[m] + Y1_ang[m, n])
                    else:
                        Jqtheta[i, k] = -V1_abs[m] * V1_abs[n] * Y1_abs[m, n] * math.cos(
                            del_1[n] - del_1[m] + Y1_ang[m, n])

            # Jqv - Derivative of Reactive Power Injections with Voltage
            Jqv = np.real(arraymXn(npq, npq))
            for i in range(npq):
                m = pq[i]
                for k in range(npq):
                    n = pq[k]
                    if n == m:
                        for n in range(nbus):
                            if m != n:
                                Jqv[i, k] = Jqv[i, k] - V1_abs[n] * Y1_abs[m, n] * math.sin(
                                    del_1[n] - del_1[m] + Y1_ang[m, n])
                        Jqv[i, k] = Jqv[i, k] - 2 * V1_abs[m] * Y1_img[m, m]
                    else:
                        Jqv[i, k] = - V1_abs[m] * Y1_abs[m, n] * math.sin(del_1[n] - del_1[m] + Y1_ang[m, n])
                        Jqv[k, i] = - V1_abs[n] * Y1_abs[m, n] * math.sin(del_1[m] - del_1[n] + Y1_ang[m, n])

            # Form the Jacobian matrix by concatenating
            J_build1 = np.hstack((Jptheta, Jpv))
            J_build2 = np.hstack((Jqtheta, Jqv))
            J = np.concatenate((J_build1, J_build2), axis=0)

            J_mx = np.asmatrix(J)
            J_inv_mx = np.linalg.inv(J_mx)
            M_mx = np.asmatrix(M)
            D = J_inv_mx * M_mx

            # Update voltage angles
            for i in range(1, nbus):
                del_1[i] = np.real(del_1[i] + D[i - 1])

            dv = nbus

            # Update only non-generator buses
            for i in range(nbus):
                if typ[i] == 1:
                    V1_abs[i] = V1_abs[i] + np.real(D[dv - 1])
                    dv = dv + 1

            # Linear formulation of negative and zero sequence voltages
            self.V0 = np.dot(Y0_inv, I0_sp)
            self.V2 = np.dot(Y2_inv, I2_sp)
            self.V0 = np.reshape(self.V0, nbus)
            self.V2 = np.reshape(self.V0, nbus)

            # Update positive sequence voltages as complex number
            for i in range(nbus):
                self.V1[i] = (V1_abs[i] * math.cos(del_1[i]) + 1j * (V1_abs[i] * math.sin(del_1[i])))

            for k in range(nl):
                self.Vseq_final[:, k] = np.array([self.V0[k], self.V1[k], self.V2[k]])

            # Convert sequence voltages to phase voltages
            self.V3ph_final = threephasetoseq(nl, k, self.T2, self.Vseq_final, self.V3ph_final)
            [self.Va, self.Vb, self.Vc] = [self.V3ph_final[0, :], self.V3ph_final[1, :], self.V3ph_final[2, :]]

            self.count = self.count + 1
            self.error = np.max(np.absolute(M))
        Iij = arraymXn(nl, nl)
        Sij = arraymXn(nl, nl)
        loss = arraymXn(nl, nl)
        S = arraymXn(nbus, 1)
        P = arraymXn(nbus, 1)
        Q = arraymXn(nbus, 1)
        susceptance = arraymXn(nl, nl)
        linechargingreac = arraymXn(nl, nl)

        for k in range(nl):
            susceptance[int(self.BranchFrom[k] - 1), int(self.BranchTo[k] - 1)] = self.s1[k]
            susceptance[int(self.BranchTo[k] - 1), int(self.BranchFrom[k] - 1)] = susceptance[int(self.BranchFrom[k] - 1), int(self.BranchTo[k] - 1)]

        for m in range(nl):
            for n in range(nl):

                linechargingreac[int(self.BranchFrom[m] - 1), int(self.BranchTo[n] - 1)] = (((V1_abs[int(self.BranchFrom[m] - 1)])**2)*susceptance[int(self.BranchFrom[m] - 1), int(self.BranchTo[n] - 1)])*self.BMva

        for k in range(nl):
            Iij[int(self.BranchFrom[k] - 1), int(self.BranchTo[k] - 1)] = -((self.V1[int(self.BranchFrom[k] - 1)] - self.V1[int(self.BranchTo[k] - 1)]) * Y1[int(self.BranchFrom[k] - 1), int(self.BranchTo[k] - 1)])
            Iij[int(self.BranchTo[k] - 1), int(self.BranchFrom[k] - 1)] = -(Iij[int(self.BranchFrom[k] - 1), int(self.BranchTo[k] - 1)])

        for m in range(nl):
            for n in range(nl):
                if m !=n:
                    Sij[m, n] = ((self.V1[m] * np.conj(Iij[m, n])) * self.BMva)-linechargingreac[m, n]
        for k in range(nl):
            loss[int(self.BranchFrom[k] - 1), int(self.BranchTo[k] - 1)] = ((np.absolute(Iij[int(self.BranchFrom[k] - 1), int(self.BranchTo[k] - 1)])**2)/(Y1[int(self.BranchFrom[k] - 1), int(self.BranchTo[k] - 1)]))*self.BMva

        for i in range(nbus):
            for j in range(nbus):
                if j != i:
                    P[i] = P[i] + V1_abs[i] * V1_abs[j] * Y1_abs[i, j] * math.cos((Y1_ang[i, j] + del_1[j] - del_1[i]))
                    Ql[i] = Ql[i] - (V1_abs[i] * V1_abs[j] * Y1_abs[i, j] * math.sin((Y1_ang[i, j] + del_1[j] - del_1[i])))

            P[i] = P[i] + V1_abs[i] * V1_abs[i] * Y1_rea[i, i]
            Q[i] = Q[i] - ((V1_abs[i] * V1_abs[i] * Y1_img[i, i]))

        for i in range(nbus):
            P[i] = P[i] - Pload[i]
            Q[i] = Q[i] - Qload[i]

        I_test = arraymXn(nbus, 1)
        S_test = arraymXn(nbus, 1)
        PF = arraymXn(nbus, 1)
        Theta = arraymXn(nbus, 1)
        Bij = arraymXn(nbus, nbus)

        for i in range(nbus):
            for j in range(nbus):
                I_test[i] = I_test[i] + (Y1[i, j]*self.V1[j])

        for i in range(nbus):
            for j in range(nbus):
                S_test[i] = S_test[i] + (Y1[i, j]*self.V1[j]*(np.conj(self.V1[i])))

        Y1inv = np.linalg.inv(np.asmatrix(Y1))
        S_test = np.conj(S_test)
        P_test = np.real(S_test)
        Q_test = np.imag(S_test)
        V1_del = np.angle(self.V1)
        V1abs = np.absolute(self.V1)
        Res = np.real(Y1inv)
        for i in range(nbus):
            PF[i] = math.atan(Q_test[i]/P_test[i])

        for i in range(nbus):
            Theta[i] = V1_del[i]-PF[i]
        Theta = np.real(Theta)
        PF = np.real(PF)

        for i in range(nbus):
            for j in range(nbus):
                Bij[i, j] = (Res[i, j]/(V1abs[i]*V1abs[j]))*((math.cos(Theta[i]-Theta[j]))/(math.sin(PF[i])*math.sin(PF[j])))
        Bij = np.real(Bij)
        PLoss = 0
        for i in range(nbus):
            for j in range(nbus):
                PLoss = PLoss + (P_test[i]*Bij[i, j]*P_test[j])

        Va_angle = (np.rad2deg(np.angle(self.Va))).astype(np.float)
        Vb_angle = (np.rad2deg(np.angle(self.Vb))).astype(np.float)
        Vc_angle = (np.rad2deg(np.angle(self.Vc))).astype(np.float)
        error_matrixtrans = np.array([Pl_a[5,], Pl_b[5,], Pl_c[5,], Ql_a[5,], Ql_b[5,], Ql_c[5,], np.absolute(self.Va)[5,], np.absolute(self.Vb)[5,], np.absolute(self.Vc)[5,]])
        Ploadbus = np.array([Pl_a[5,], Pl_b[5,], Pl_c[5,]])
        Qloadbus = np.array([Ql_a[5,], Ql_b[5,], Ql_c[5,]])
        d = {'Va': (((np.absolute(self.Va)).astype(np.float))),
             'Vb': (((np.absolute(self.Vb)).astype(np.float))),
             'Vc': (((np.absolute(self.Vc)).astype(np.float))),
             'Va_angle': Va_angle,
             'Power_Flows': Sij,
             'Vb_angle': Vb_angle,
             'Vc_angle': Vc_angle,
             'Ploadbus': Ploadbus,
             'Qloadbus': Qloadbus,
             'error_matrixtrans': error_matrixtrans
             }
        return d

    def main(self):
        a = SDTPF(9)
        data = a.compiledata()
        result = a.threeseq_solve(*data)

if __name__ == '__main__':
    SDTPF.main('transmission')

